@@ -2,7 +2,7 @@
  *
  *  Connection Manager
  *
- *  Copyright (C) 2007-2009  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2007-2010  Intel Corporation. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -34,6 +34,7 @@
 #include <connman/plugin.h>
 #include <connman/resolver.h>
 #include <connman/notifier.h>
+#include <connman/ondemand.h>
 #include <connman/log.h>
 
 #include <glib.h>
@@ -92,12 +93,16 @@ struct request_data {
 	guint timeout;
 	guint numserv;
 	guint numresp;
+	gpointer request;
+	gsize request_len;
+	gpointer name;
 	gpointer resp;
 	gsize resplen;
 };
 
 static GSList *server_list = NULL;
 static GSList *request_list = NULL;
+static GSList *request_pending_list = NULL;
 static guint16 request_id = 0x0000;
 
 static GIOChannel *listener_channel = NULL;
@@ -297,207 +302,6 @@ static void destroy_server(struct server_data *data)
 	g_free(data);
 }
 
-static int dnsproxy_append(const char *interface, const char *domain,
-							const char *server)
-{
-	struct server_data *data;
-
-	DBG("interface %s server %s", interface, server);
-
-	if (g_str_equal(server, "127.0.0.1") == TRUE)
-		return -ENODEV;
-
-	data = create_server(interface, domain, server);
-	if (data == NULL)
-		return -EIO;
-
-	server_list = g_slist_append(server_list, data);
-
-	return 0;
-}
-
-static int dnsproxy_remove(const char *interface, const char *domain,
-							const char *server)
-{
-	struct server_data *data;
-
-	DBG("interface %s server %s", interface, server);
-
-	if (g_str_equal(server, "127.0.0.1") == TRUE)
-		return -ENODEV;
-
-	data = find_server(interface, domain, server);
-	if (data == NULL)
-		return 0;
-
-	server_list = g_slist_remove(server_list, data);
-
-	destroy_server(data);
-
-	return 0;
-}
-
-static struct connman_resolver dnsproxy_resolver = {
-	.name		= "dnsproxy",
-	.priority	= CONNMAN_RESOLVER_PRIORITY_HIGH,
-	.append		= dnsproxy_append,
-	.remove		= dnsproxy_remove,
-};
-
-static void dnsproxy_offline_mode(connman_bool_t enabled)
-{
-	GSList *list;
-
-	DBG("enabled %d", enabled);
-
-	for (list = server_list; list; list = list->next) {
-		struct server_data *data = list->data;
-
-		if (enabled == FALSE) {
-			connman_info("Enabling DNS server %s", data->server);
-			data->enabled = TRUE;
-		} else {
-			connman_info("Disabling DNS server %s", data->server);
-			data->enabled = FALSE;
-		}
-	}
-}
-
-static void dnsproxy_default_changed(struct connman_service *service)
-{
-	GSList *list;
-	char *interface;
-
-	DBG("service %p", service);
-
-	if (service == NULL) {
-		/* When no services are active, then disable DNS proxying */
-		dnsproxy_offline_mode(TRUE);
-		return;
-	}
-
-	interface = connman_service_get_interface(service);
-	if (interface == NULL)
-		return;
-
-	for (list = server_list; list; list = list->next) {
-		struct server_data *data = list->data;
-
-		if (g_strcmp0(data->interface, interface) == 0) {
-			connman_info("Enabling DNS server %s", data->server);
-			data->enabled = TRUE;
-		} else {
-			connman_info("Disabling DNS server %s", data->server);
-			data->enabled = FALSE;
-		}
-	}
-
-	g_free(interface);
-}
-
-static struct connman_notifier dnsproxy_notifier = {
-	.name			= "dnsproxy",
-	.default_changed	= dnsproxy_default_changed,
-	.offline_mode		= dnsproxy_offline_mode,
-};
-
-static int parse_request(unsigned char *buf, int len,
-					char *name, unsigned int size)
-{
-	struct domain_hdr *hdr = (void *) buf;
-	uint16_t qdcount = ntohs(hdr->qdcount);
-	uint16_t arcount = ntohs(hdr->arcount);
-	unsigned char *ptr;
-	char *last_label = NULL;
-	int label_count = 0;
-	unsigned int remain, used = 0;
-
-	if (len < 12)
-		return -EINVAL;
-
-	DBG("id 0x%04x qr %d opcode %d qdcount %d arcount %d",
-					hdr->id, hdr->qr, hdr->opcode,
-							qdcount, arcount);
-
-	if (hdr->qr != 0 || qdcount != 1)
-		return -EINVAL;
-
-	memset(name, 0, size);
-
-	ptr = buf + 12;
-	remain = len - 12;
-
-	while (remain > 0) {
-		uint8_t len = *ptr;
-
-		if (len == 0x00) {
-			if (label_count > 0)
-				last_label = (char *) (ptr + 1);
-			break;
-		}
-
-		label_count++;
-
-		if (used + len + 1 > size)
-			return -ENOBUFS;
-
-		strncat(name, (char *) (ptr + 1), len);
-		strcat(name, ".");
-
-		used += len + 1;
-
-		ptr += len + 1;
-		remain -= len + 1;
-	}
-
-	if (arcount && remain >= 9 && last_label[4] == 0 &&
-				last_label[5] == 0 && last_label[6] == 0x29) {
-		uint16_t edns0_bufsize;
-
-		edns0_bufsize = last_label[7] << 8 | last_label[8];
-
-		DBG("EDNS0 buffer size %u", edns0_bufsize);
-
-		/* This is an evil hack until full TCP support has been
-		 * implemented.
-		 *
-		 * Somtimes the EDNS0 request gets send with a too-small
-		 * buffer size. Since glibc doesn't seem to crash when it
-		 * gets a response biffer then it requested, just bump
-		 * the buffer size up to 4KiB.
-		 */
-		if (edns0_bufsize < 0x1000) {
-			last_label[7] = 0x10;
-			last_label[8] = 0x00;
-		}
-	}
-
-	DBG("query %s (%d labels)", name, label_count);
-
-	return 0;
-}
-
-static void send_response(int sk, unsigned char *buf, int len,
-				const struct sockaddr *to, socklen_t tolen)
-{
-	struct domain_hdr *hdr = (void *) buf;
-	int err;
-
-	if (len < 12)
-		return;
-
-	DBG("id 0x%04x qr %d opcode %d", hdr->id, hdr->qr, hdr->opcode);
-
-	hdr->qr = 1;
-	hdr->rcode = 2;
-
-	hdr->ancount = 0;
-	hdr->nscount = 0;
-	hdr->arcount = 0;
-
-	err = sendto(sk, buf, len, 0, to, tolen);
-}
-
 static int append_query(unsigned char *buf, unsigned int size,
 				const char *query, const char *domain)
 {
@@ -576,10 +380,287 @@ static gboolean request_timeout(gpointer user_data)
 	return FALSE;
 }
 
+static gboolean resolv(struct request_data *req,
+				gpointer request, gpointer name)
+{
+	int sk, err;
+	GSList *list;
+
+	request_list = g_slist_append(request_list, req);
+
+	req->numserv = 0;
+	req->timeout = g_timeout_add_seconds(5, request_timeout, req);
+
+	for (list = server_list; list; list = list->next) {
+		struct server_data *data = list->data;
+
+		DBG("server %s domain %s enabled %d",
+				data->server, data->domain, data->enabled);
+
+		if (data->enabled == FALSE)
+			continue;
+
+		sk = g_io_channel_unix_get_fd(data->channel);
+
+		err = send(sk, request, req->request_len, 0);
+
+		req->numserv++;
+
+		if (data->domain != NULL) {
+			unsigned char alt[1024];
+			struct domain_hdr *hdr = (void *) &alt;
+			int altlen, domlen;
+
+			domlen = strlen(data->domain) + 1;
+			if (domlen < 5)
+				continue;
+
+			alt[0] = req->altid & 0xff;
+			alt[1] = req->altid >> 8;
+
+			memcpy(alt + 2, request + 2, 10);
+			hdr->qdcount = htons(1);
+
+			altlen = append_query(alt + 12, sizeof(alt) - 12,
+						name, data->domain);
+			if (altlen < 0)
+				continue;
+
+			altlen += 12;
+
+			memcpy(alt + altlen, request + altlen - domlen,
+					req->request_len - altlen + domlen);
+
+			err = send(sk, alt, req->request_len + domlen + 1, 0);
+
+			req->numserv++;
+		}
+	}
+
+	return TRUE;
+}
+
+static int dnsproxy_append(const char *interface, const char *domain,
+							const char *server)
+{
+	struct server_data *data;
+
+	DBG("interface %s server %s", interface, server);
+
+	if (g_str_equal(server, "127.0.0.1") == TRUE)
+		return -ENODEV;
+
+	data = create_server(interface, domain, server);
+	if (data == NULL)
+		return -EIO;
+
+	server_list = g_slist_append(server_list, data);
+
+	return 0;
+}
+
+static int dnsproxy_remove(const char *interface, const char *domain,
+							const char *server)
+{
+	struct server_data *data;
+
+	DBG("interface %s server %s", interface, server);
+
+	if (g_str_equal(server, "127.0.0.1") == TRUE)
+		return -ENODEV;
+
+	data = find_server(interface, domain, server);
+	if (data == NULL)
+		return 0;
+
+	server_list = g_slist_remove(server_list, data);
+
+	destroy_server(data);
+
+	return 0;
+}
+
+static void dnsproxy_flush(void)
+{
+	GSList *list;
+
+	list = request_pending_list;
+	while (list) {
+		struct request_data *req = list->data;
+
+		list = list->next;
+
+		request_pending_list =
+				g_slist_remove(request_pending_list, req);
+		resolv(req, req->request, req->name);
+		g_free(req->request);
+		g_free(req->name);
+	}
+}
+
+static struct connman_resolver dnsproxy_resolver = {
+	.name		= "dnsproxy",
+	.priority	= CONNMAN_RESOLVER_PRIORITY_HIGH,
+	.append		= dnsproxy_append,
+	.remove		= dnsproxy_remove,
+	.flush		= dnsproxy_flush,
+};
+
+static void dnsproxy_offline_mode(connman_bool_t enabled)
+{
+	GSList *list;
+
+	DBG("enabled %d", enabled);
+
+	for (list = server_list; list; list = list->next) {
+		struct server_data *data = list->data;
+
+		if (enabled == FALSE) {
+			connman_info("Enabling DNS server %s", data->server);
+			data->enabled = TRUE;
+		} else {
+			connman_info("Disabling DNS server %s", data->server);
+			data->enabled = FALSE;
+		}
+	}
+}
+
+static void dnsproxy_default_changed(struct connman_service *service)
+{
+	GSList *list;
+	char *interface;
+
+	DBG("service %p", service);
+
+	if (service == NULL) {
+		/* When no services are active, then disable DNS proxying */
+		dnsproxy_offline_mode(TRUE);
+		return;
+	}
+
+	interface = connman_service_get_interface(service);
+	if (interface == NULL)
+		return;
+
+	for (list = server_list; list; list = list->next) {
+		struct server_data *data = list->data;
+
+		if (g_strcmp0(data->interface, interface) == 0) {
+			connman_info("Enabling DNS server %s", data->server);
+			data->enabled = TRUE;
+		} else {
+			connman_info("Disabling DNS server %s", data->server);
+			data->enabled = FALSE;
+		}
+	}
+
+	g_free(interface);
+}
+
+static struct connman_notifier dnsproxy_notifier = {
+	.name			= "dnsproxy",
+	.default_changed	= dnsproxy_default_changed,
+	.offline_mode		= dnsproxy_offline_mode,
+};
+
+static unsigned char opt_edns0_type[2] = { 0x00, 0x29 };
+
+static int parse_request(unsigned char *buf, int len,
+					char *name, unsigned int size)
+{
+	struct domain_hdr *hdr = (void *) buf;
+	uint16_t qdcount = ntohs(hdr->qdcount);
+	uint16_t arcount = ntohs(hdr->arcount);
+	unsigned char *ptr;
+	char *last_label = NULL;
+	unsigned int remain, used = 0;
+
+	if (len < 12)
+		return -EINVAL;
+
+	DBG("id 0x%04x qr %d opcode %d qdcount %d arcount %d",
+					hdr->id, hdr->qr, hdr->opcode,
+							qdcount, arcount);
+
+	if (hdr->qr != 0 || qdcount != 1)
+		return -EINVAL;
+
+	memset(name, 0, size);
+
+	ptr = buf + sizeof(struct domain_hdr);
+	remain = len - sizeof(struct domain_hdr);
+
+	while (remain > 0) {
+		uint8_t len = *ptr;
+
+		if (len == 0x00) {
+			last_label = (char *) (ptr + 1);
+			break;
+		}
+
+		if (used + len + 1 > size)
+			return -ENOBUFS;
+
+		strncat(name, (char *) (ptr + 1), len);
+		strcat(name, ".");
+
+		used += len + 1;
+
+		ptr += len + 1;
+		remain -= len + 1;
+	}
+
+	if (last_label && arcount && remain >= 9 && last_label[4] == 0 &&
+				!memcmp(last_label + 5, opt_edns0_type, 2)) {
+		uint16_t edns0_bufsize;
+
+		edns0_bufsize = last_label[7] << 8 | last_label[8];
+
+		DBG("EDNS0 buffer size %u", edns0_bufsize);
+
+		/* This is an evil hack until full TCP support has been
+		 * implemented.
+		 *
+		 * Somtimes the EDNS0 request gets send with a too-small
+		 * buffer size. Since glibc doesn't seem to crash when it
+		 * gets a response biffer then it requested, just bump
+		 * the buffer size up to 4KiB.
+		 */
+		if (edns0_bufsize < 0x1000) {
+			last_label[7] = 0x10;
+			last_label[8] = 0x00;
+		}
+	}
+
+	DBG("query %s", name);
+
+	return 0;
+}
+
+static void send_response(int sk, unsigned char *buf, int len,
+				const struct sockaddr *to, socklen_t tolen)
+{
+	struct domain_hdr *hdr = (void *) buf;
+	int err;
+
+	if (len < 12)
+		return;
+
+	DBG("id 0x%04x qr %d opcode %d", hdr->id, hdr->qr, hdr->opcode);
+
+	hdr->qr = 1;
+	hdr->rcode = 2;
+
+	hdr->ancount = 0;
+	hdr->nscount = 0;
+	hdr->arcount = 0;
+
+	err = sendto(sk, buf, len, 0, to, tolen);
+}
+
 static gboolean listener_event(GIOChannel *channel, GIOCondition condition,
 							gpointer user_data)
 {
-	GSList *list;
 	unsigned char buf[768];
 	char query[512];
 	struct request_data *req;
@@ -604,7 +685,8 @@ static gboolean listener_event(GIOChannel *channel, GIOCondition condition,
 	DBG("Received %d bytes (id 0x%04x)", len, buf[0] | buf[1] << 8);
 
 	err = parse_request(buf, len, query, sizeof(query));
-	if (err < 0 || g_slist_length(server_list) == 0) {
+	if (err < 0 || (g_slist_length(server_list) == 0 &&
+				connman_ondemand_connected())) {
 		send_response(sk, buf, len, (struct sockaddr *) &sin, size);
 		return TRUE;
 	}
@@ -623,62 +705,39 @@ static gboolean listener_event(GIOChannel *channel, GIOCondition condition,
 	req->srcid = buf[0] | (buf[1] << 8);
 	req->dstid = request_id;
 	req->altid = request_id + 1;
+	req->request_len = len;
 
 	buf[0] = req->dstid & 0xff;
 	buf[1] = req->dstid >> 8;
 
-	request_list = g_slist_append(request_list, req);
+	if (!connman_ondemand_connected()) {
+		DBG("Starting on demand connection");
+		/*
+		 * We're not connected, let's queue the request and start
+		 * an on-demand connection.
+		 */
+		req->request = g_try_malloc0(req->request_len);
+		if (req->request == NULL)
+			return TRUE;
 
-	req->numserv = 0;
-	req->timeout = g_timeout_add_seconds(5, request_timeout, req);
+		memcpy(req->request, buf, req->request_len);
 
-	for (list = server_list; list; list = list->next) {
-		struct server_data *data = list->data;
-
-		DBG("server %s domain %s enabled %d",
-				data->server, data->domain, data->enabled);
-
-		if (data->enabled == FALSE)
-			continue;
-
-		sk = g_io_channel_unix_get_fd(data->channel);
-
-		err = send(sk, buf, len, 0);
-
-		req->numserv++;
-
-		if (data->domain != NULL) {
-			unsigned char alt[1024];
-			struct domain_hdr *hdr = (void *) &alt;
-			int altlen, domlen;
-
-			domlen = strlen(data->domain) + 1;
-			if (domlen < 5)
-				continue;
-
-			alt[0] = req->altid & 0xff;
-			alt[1] = req->altid >> 8;
-
-			memcpy(alt + 2, buf + 2, 10);
-			hdr->qdcount = htons(1);
-
-			altlen = append_query(alt + 12, sizeof(alt) - 12,
-							query, data->domain);
-			if (altlen < 0)
-				continue;
-
-			altlen += 12;
-
-			memcpy(alt + altlen, buf + altlen - domlen,
-							len - altlen + domlen);
-
-			err = send(sk, alt, len + domlen + 1, 0);
-
-			req->numserv++;
+		req->name = g_try_malloc0(sizeof(query));
+		if (req->name == NULL) {
+			g_free(req->request);
+			return TRUE;
 		}
+		memcpy(req->name, query, sizeof(query));
+
+		request_pending_list = g_slist_append(request_pending_list,
+									req);
+
+		connman_ondemand_start("", 300);
+
+		return TRUE;
 	}
 
-	return TRUE;
+	return resolv(req, buf, query);
 }
 
 static int create_listener(void)
@@ -745,6 +804,22 @@ static void destroy_listener(void)
 	if (listener_watch > 0)
 		g_source_remove(listener_watch);
 
+	for (list = request_pending_list; list; list = list->next) {
+		struct request_data *req = list->data;
+
+		DBG("Dropping pending request (id 0x%04x -> 0x%04x)",
+						req->srcid, req->dstid);
+
+		g_free(req->resp);
+		g_free(req->request);
+		g_free(req->name);
+		g_free(req);
+		list->data = NULL;
+	}
+
+	g_slist_free(request_pending_list);
+	request_pending_list = NULL;
+
 	for (list = request_list; list; list = list->next) {
 		struct request_data *req = list->data;
 
@@ -752,6 +827,8 @@ static void destroy_listener(void)
 						req->srcid, req->dstid);
 
 		g_free(req->resp);
+		g_free(req->request);
+		g_free(req->name);
 		g_free(req);
 		list->data = NULL;
 	}
